#include "sdk_config.h"

#ifdef TEST
#include "nrf.h"
#include "nrf_drv_clock.h"
#include "nrf_gpio.h"
#include "nrf_delay.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_drv_power.h"
#include "nrf_serial.h"
#include "app_timer.h"

#include "app_error.h"
#include "app_util.h"
#include "boards.h"

#include "thread_coap_utils.h"
#include "uart.h"
#include "uart_interpreter.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include "uart.h"
#include "uart_interpreter.h"
#include "boards.h"
#include "nrf.h"
#include "nrf_drv_clock.h"
#include "nrf_gpio.h"
#include "nrf_delay.h"
#include "nrf_log.h"
#include "nrf_drv_power.h"
#include "thread_coap_utils.h"

//Ble UART:
#if defined (UART_PRESENT)
    #include "nrf_uart.h"
#endif
#if defined (UARTE_PRESENT)
    #include "nrf_uarte.h"
#endif

#if NRF_MODULE_ENABLED(NRF_SERIAL)
  #include "nrf_drv_common.h"
#endif
#endif


#define COMMAND_LED4         "LED4"



NRF_SERIAL_UART_DEF(serial0_uarte, 1);
NRF_SERIAL_UART_DEF(serial1_uarte, 0);

static void sleep_handler(void)
{
    __WFE();
    __SEV();
    __WFE();
}


/**@brief   Function for handling uart events.
 *
 * @details This function will receive a single character from the uart module and append it to
 *          a string. The string will be be sent over BLE when the last character received was a
 *          'new line' '\n' (hex 0x0A) or if the string has reached the maximum data length.
 */
/**@snippet [Handling the data received over UART] */
static void uart_event_handle(struct nrf_serial_s const *p_serial, nrf_serial_event_t event)
{
    //static uint8_t buffer[UART_RX_BUF_SIZE];
    static uint8_t index = 0;
    uint32_t err_code = NRF_SUCCESS;
    ret_code_t ret;
    char c;

    switch (event)
    {
        case NRF_SERIAL_EVENT_TX_DONE:
            break;

        case NRF_SERIAL_EVENT_RX_DATA:
            //ret = nrf_serial_read(p_serial, &buffer, sizeof(buffer), &nbr_data_read, 0);
            ret = nrf_queue_read(p_serial->p_ctx->p_config->p_queues->p_rxq, &c, sizeof(c));
            uart_interpreter_process_data_received(c);
            /*m_frame.p_payload[index] = c;
            if (c == '\n')
            {
                m_frame.payload_length = index+1;

                err_code = app_sched_event_put(&m_frame, sizeof(m_frame), process_received_frame);
                ASSERT(err_code == NRF_SUCCESS);

                index = 0; 
                break;
            }

            if(index >= (UART_RX_BUF_SIZE-1))
            {
                index = 0; 
                nrf_serial_rx_drain(&serial0_uarte);
                break;
            }
            else
            {
                index++;
            }*/
            break;

        case NRF_SERIAL_EVENT_DRV_ERR :
            //APP_ERROR_HANDLER(p_event->data.error_communication);
            uart_log_send_string("UART: Internal driver error");
            break;

        case NRF_SERIAL_EVENT_FIFO_ERR :
            //APP_ERROR_HANDLER(p_event->data.error_code);
            uart_log_send_string("UART: RX FIFO overrun");
            break;

        default:
            uart_log_send_string("UART: Unknown error");
            break;
    }
}
/**@snippet [Handling the data received over UART] */

#define SERIAL_FIFO_TX_SIZE 32
#define SERIAL_FIFO_RX_SIZE 32
#define SERIAL_BUFF_TX_SIZE 1
#define SERIAL_BUFF_RX_SIZE 1

//Main UART:
NRF_SERIAL_DRV_UART_CONFIG_DEF(m_uarte0_drv_config,
                      RX_PIN_NUMBER, TX_PIN_NUMBER,
                      RTS_PIN_NUMBER, CTS_PIN_NUMBER,
                      NRF_UART_HWFC_DISABLED, NRF_UART_PARITY_EXCLUDED,
                      NRF_UART_BAUDRATE_115200,
                      UART_DEFAULT_CONFIG_IRQ_PRIORITY);
NRF_SERIAL_QUEUES_DEF(serial0_queues, SERIAL_FIFO_TX_SIZE, SERIAL_FIFO_RX_SIZE);
NRF_SERIAL_BUFFERS_DEF(serial0_buffs, SERIAL_BUFF_TX_SIZE, SERIAL_BUFF_RX_SIZE);
NRF_SERIAL_CONFIG_DEF(serial0_config, NRF_SERIAL_MODE_IRQ,
                      &serial0_queues, &serial0_buffs, uart_event_handle, sleep_handler);
//NRF_SERIAL_UART_DEF(serial0_uarte, 1);


//Logs UART:
NRF_SERIAL_DRV_UART_CONFIG_DEF(m_uarte1_drv_config,
                      NRF_UART_PSEL_DISCONNECTED, NRF_LOG_BACKEND_UART_TX_PIN,
                      NULL, NULL,
                      NRF_UART_HWFC_DISABLED, NRF_UART_PARITY_EXCLUDED,
                      NRF_UART_BAUDRATE_115200,
                      UART_DEFAULT_CONFIG_IRQ_PRIORITY);
NRF_SERIAL_QUEUES_DEF(serial1_queues, SERIAL_FIFO_TX_SIZE, SERIAL_FIFO_RX_SIZE);
NRF_SERIAL_BUFFERS_DEF(serial1_buffs, SERIAL_BUFF_TX_SIZE, SERIAL_BUFF_RX_SIZE);
NRF_SERIAL_CONFIG_DEF(serial1_config, NRF_SERIAL_MODE_POLLING,
                      &serial1_queues, &serial1_buffs, NULL, sleep_handler);
//NRF_SERIAL_UART_DEF(serial1_uarte, 0);


/**@brief  Function for initializing the UART module.
 */
/**@snippet [UART Initialization] */
void uart_init(void)
{
      ret_code_t ret;

      ret = nrf_drv_clock_init();
      APP_ERROR_CHECK(ret);
      ret = nrf_drv_power_init(NULL);
      APP_ERROR_CHECK(ret);

      nrf_drv_clock_lfclk_request(NULL);
      ret = app_timer_init();
      APP_ERROR_CHECK(ret);

      ret = nrf_serial_init(&serial0_uarte, &m_uarte0_drv_config, &serial0_config); // Main UART
      APP_ERROR_CHECK(ret);

      ret = nrf_serial_init(&serial1_uarte, &m_uarte1_drv_config, &serial1_config);  // Logs
      APP_ERROR_CHECK(ret);
}
/**@snippet [UART Initialization] */

/***************************************************************************************************
 * @section Command parsing
 **************************************************************************************************/
//To do : decode the data in uart_interpreter and call ble_command_handler in main
void ble_command_handler(const uint8_t * p_command_str, uint16_t length) // Decode the received data from BLE
{
    /*for(int i=0; i<length; i++)
    {
//        uart_interpreter_process_data_received(p_command_str[i]);
    }*/

    if (strncmp(COMMAND_LED4, (char *)p_command_str, 4) == 0)
    {
        uint8_t command = 0;

        if(strstr((char *)p_command_str, "ON") != NULL)
        {
            command = THREAD_COAP_UTILS_LIGHT_CMD_ON;
        }
        else if(strstr((char *)p_command_str, "OFF") != NULL)
        {
            command = THREAD_COAP_UTILS_LIGHT_CMD_OFF;
        }
        else
        {
          //Command error
        }

        thread_coap_utils_change_light(command);

        if(command != 0)
        {
            thread_coap_utils_multicast_light_request_send(command,
                                                           THREAD_COAP_UTILS_MULTICAST_REALM_LOCAL);            
        }
    }
    /*else if (strncmp(COMMAND_SEND, (char *)p_command_str, 4) == 0)
    {

    }
    else if (p_command_str[0] == 0xAA) //[Sync Flag = 0xAA] [nbr bytes] [Payload]
    {
        if(length>2)
        {
            uint8_t payloadLength = p_command_str[1]; 
            uint8_t cmd = p_command_str[2]; 
            uint32_t err_code = 0;
            const char  message_unknownCmd[] = "Unknown cmd received\r\n";
            const char  message_test[] = "Test\r\n";
            ret_code_t ret;
            
            switch(cmd)
            {
                case 1:  // Test cmd
                    ret = uart_log_send_string(message_test);
                    APP_ERROR_CHECK(ret);
                    break;

                case 2:  // Get neighbors informations
get_thread_neighbors_list();
                    break;

                default:
                    ret = uart_log_send_string(message_unknownCmd);
                    APP_ERROR_CHECK(ret);
                    break;
            }
        }
    }*/
    else //Send the received frame to the thread network
    {
        char *uart_cmd = (char *)malloc(length+1);
        strcpy(uart_cmd, p_command_str);
        thread_coap_utils_multicast_uart_request_send(uart_cmd,
                                                       length,
                                                       THREAD_COAP_UTILS_MULTICAST_REALM_LOCAL);
    }
    
}



ret_code_t uart_main_send_string(const char *str)
{
	ret_code_t  ret;

        ret = nrf_serial_write(&serial0_uarte, str, strlen(str), NULL, NRF_SERIAL_MAX_TIMEOUT);
        (void)nrf_serial_flush(&serial0_uarte, 0);

	return ret;
}

ret_code_t uart_main_send_bytesArray(uint8_t data_array[], uint8_t nbr_bytes)
{
	ret_code_t  ret;

        ret = nrf_serial_write(&serial0_uarte, data_array, nbr_bytes, NULL, NRF_SERIAL_MAX_TIMEOUT);
        (void)nrf_serial_flush(&serial0_uarte, 0);

	return ret;
}

ret_code_t uart_log_send_string(const char *str)
{
	ret_code_t  ret;

        ret = nrf_serial_write(&serial1_uarte, str, strlen(str), NULL, NRF_SERIAL_MAX_TIMEOUT);
        ret = nrf_serial_write(&serial1_uarte, &"\r\n", 2, NULL, NRF_SERIAL_MAX_TIMEOUT);
        (void)nrf_serial_flush(&serial1_uarte, 0);

	return ret;
}

ret_code_t uart_log_send_stringWithFrame(const char *str, uint8_t data_array[], uint8_t nbr_bytes)
{
        ret_code_t  ret;

        ret = nrf_serial_write(&serial1_uarte, str, strlen(str), NULL, NRF_SERIAL_MAX_TIMEOUT);

        for(int j=0; j<nbr_bytes; j++)
        {
            char ascii_buffer[3];
            utoa (data_array[j], ascii_buffer, 16);
            if(data_array[j]<=0x0F)
            {
              nrf_serial_write(&serial1_uarte, &"0", 1, NULL, NRF_SERIAL_MAX_TIMEOUT);
              nrf_serial_write(&serial1_uarte, &ascii_buffer, 1, NULL, NRF_SERIAL_MAX_TIMEOUT);
            }
            else
            {
              nrf_serial_write(&serial1_uarte, &ascii_buffer, 2, NULL, NRF_SERIAL_MAX_TIMEOUT);
            }
            nrf_serial_write(&serial1_uarte, &" ", 1, NULL, NRF_SERIAL_MAX_TIMEOUT);
        }

        ret = nrf_serial_write(&serial1_uarte, &"\r\n", 2, NULL, NRF_SERIAL_MAX_TIMEOUT);
        (void)nrf_serial_flush(&serial1_uarte, 0);

	return ret;
}

